import java.io.*;
import java.util.Vector;
import java.util.Enumeration;

class TableEntry {
    // Overview: Record-like class for implementing the StringTable.

    /*@non_null@*/ private String name;
    private double rate;
    
    public TableEntry (/*@non_null@*/ String n, double r)
    {
	name = n;
	rate = r;
    }
    
    public String getName () 
    //@ensures \result != null
    {
	return name;
    }
    
    public double getRate ()
    {
	return rate;
    }
    
    public String toString ()
    {
	return name + ": " + rate;
    }
}

public class StringTable 
{
    private /*@non_null@*/ Vector entries ;

    //@ghost public int numEntries ; 
    //@invariant entries.elementType == \type(TableEntry) ;
    //@invariant entries.containsNull == false ; 
    //@invariant entries.elementCount == numEntries ;

    // invariant the entries are sorted by their values.
    
    public StringTable () 
    //@ensures numEntries == 0;
    {
	entries = new Vector ();
    } /* ESC/Java cannot establish invariant for empty Vector */ //@nowarn Invariant
    
    // Read a StringTable structure from a file formated as the output of toString
    public StringTable (InputStream instream) 
    {
	this ();
	
	try {
	    StructuredReader reader = new StructuredReader (instream);
	    while (true) {
		try {
		    String name = reader.readThroughAny (":");
		    reader.skipWhitespace ();
		    double rate = reader.readDouble ();
		    reader.skipWhitespace ();
		    addName (name, rate);
		} catch (EOFException e) {
		    break;
		} 
	    } 
	    
	    reader.close ();
	    // It would be better to propagate these exceptions to the caller, but for PS2, we 
	    // avoid the use of exceptions.
	} catch (DuplicateEntryException e) {
	    Assert.badBranch ("Names file contains duplicate entry: " + e);
	} catch (IOException e) {
	    Assert.badBranch ("Error reading names file: " + e);
	} catch (NoNumberException e) {
	    Assert.badBranch ("Error reading number in names file: " + e);
	} catch (NumberFormatException e) {
	    Assert.badBranch ("Number format error reading names file: " + e);
	}
    }
    
    private TableEntry lookupEntry (String name)
    {
	for (Enumeration e = entries.elements (); e.hasMoreElements (); ) {
	    TableEntry entry = (TableEntry) e.nextElement ();
	    if (entry.getName ().equals (name)) {
		return entry;
	    }
	}
	
	return null;
    }
    
    public void addName (/*@non_null@*/ String name, double rate) throws DuplicateEntryException
    //@modifies numEntries;
    //@ensures numEntries >= \old(numEntries);
    {
	if (lookupEntry (name) != null) {
	    throw new DuplicateEntryException (name);
	}
	
	// Linear search to find the location to insert (before the first element greater than this one)
	int index;
	
	for (index = 1; index <= entries.size (); index++) {
	    TableEntry entry = (TableEntry) entries.elementAt (index - 1);
	    if (entry.getRate () > rate) {
		break;
	    }
	}
	
	index--;
	entries.insertElementAt (new TableEntry (name, rate), index);
    }
    
    public double getRate (String name)
    {
	TableEntry entry = lookupEntry (name);
	
	if (entry != null) {
	    return entry.getRate ();
	} else {
	    return 0;
	}
    }
    
    public /*@non_null@*/ String getNthLowest (int index)
       //@requires index >= 0;
       //@requires index < numEntries;
    {
	TableEntry entry = (TableEntry) entries.elementAt (index);
	return entry.getName ();
    }
    
    public int size ()
       //@ensures \result == numEntries;
       // effects: Returns the number of distinct names.
    {
	return entries.size ();
    }
    
    public /*@non_null@*/ StringIterator keys () 
    {
	Vector allNames = new java.util.Vector ();
	//@set allNames.elementType = \type(String) ;
	//@set allNames.containsNull = false ;
	
	for (Enumeration e = entries.elements (); e.hasMoreElements (); ) {
	    allNames.addElement (((TableEntry) (e.nextElement ())).getName ());
	}
	
	return new StringIterator (allNames.elements ());
    } /*@nowarn Invariant@*/ /* ESC/Java has a bug in checking the invariant here */
    
    public String toString ()
    {
	StringBuffer res = new StringBuffer ();
	
	for (int index = entries.size () - 1; index >= 0; index--) {
	    TableEntry entry = (TableEntry) entries.elementAt (index);
	    res.append (entry.toString ());
	    res.append ('\n');
	}
	
	return new String (res);
    }
}









